@@ -422,7 +422,7 @@ class Client:
         Warning:
             This is not finished, and the API may change at any point!
         """
-        at = datetime.datetime.utcnow()
+        at = _util.now()
         form = {
             "folders[0]": "inbox",
             "client": "mercury",
