@@ -21,10 +21,9 @@ from . import _graphql, _util, _exception
 from typing import Optional, Mapping, Callable, Any, Awaitable
 
 SERVER_JS_DEFINE_REGEX = re.compile(r'(?:'
-                                    r'\(new ServerJS\(\)\)'
-                                    r'|\(require\("ServerJS"\)\)\(\)'
-                                    r'|\(new ServerJS\(\)\);s'
-                                    r').handle\(')
+                                    r'\(new ServerJS\(\)\)(?:;s)?'
+                                    r'|\(require\("ServerJS(?:Define)?"\)\)\(\)'
+                                    r').handle(?:Defines)?\(')
 SERVER_JS_DEFINE_JSON_DECODER = json.JSONDecoder()
 
 
@@ -66,11 +65,13 @@ def parse_server_js_define(html: str) -> Mapping[str, Any]:
         rtn = parsed["define"]
     except KeyError:
         file_name = write_html_to_temp(html)
-        raise _exception.ParseError("Invalid ServerJSDefine: missing define key", data_file=file_name)
+        raise _exception.ParseError("Invalid ServerJSDefine: missing define key",
+                                    data_file=file_name)
 
     if not isinstance(rtn, list):
         file_name = write_html_to_temp(html)
-        raise _exception.ParseError("Invalid ServerJSDefine: define value is not a list", data_file=file_name)
+        raise _exception.ParseError("Invalid ServerJSDefine: define value is not a list",
+                                    data_file=file_name)
 
     # Convert to a dict
     return _util.get_jsmods_define(rtn)
@@ -114,17 +115,16 @@ def get_user_id(session: aiohttp.ClientSession) -> str:
     return rtn if isinstance(rtn, str) else str(rtn.value)
 
 
-def session_factory() -> aiohttp.ClientSession:
+def session_factory(user_agent: Optional[str] = None) -> aiohttp.ClientSession:
     from . import __version__
-
-    session = aiohttp.ClientSession(headers={
+    return aiohttp.ClientSession(headers={
         "Referer": "https://www.messenger.com/",
-        # We won't try to set a fake user agent to mask our presence!
-        # Facebook allows us access anyhow, and it makes our motives clearer:
-        # We're not trying to cheat Facebook, we simply want to access their service
-        "User-Agent": "fbchat-asyncio/{}".format(__version__)
+        "User-Agent": user_agent or f"fbchat-asyncio/{__version__}",
     })
-    return session
+
+
+def login_cookies(at: datetime.datetime):
+    return {"act": "{}/0".format(_util.datetime_to_millis(at))}
 
 
 def client_id_factory() -> str:
@@ -164,7 +164,8 @@ async def two_factor_helper(session: aiohttp.ClientSession, r: aiohttp.ClientRes
     while "approvals_code" in data:
         data["approvals_code"] = await on_2fa_callback()
         log.info("Submitting 2FA code")
-        r = await session.post(url, data=data, allow_redirects=False)
+        r = await session.post(url, data=data, allow_redirects=False,
+                               cookies=login_cookies(_util.now()))
         log.debug("2FA location: %s", r.headers.get("Location"))
         url, data = find_form_request(await r.text())
 
@@ -172,22 +173,32 @@ async def two_factor_helper(session: aiohttp.ClientSession, r: aiohttp.ClientRes
     if "name_action_selected" in data:
         data["name_action_selected"] = "save_device"
         log.info("Saving browser")
-        r = await session.post(url, data=data, allow_redirects=False)
+        r = await session.post(url, data=data, allow_redirects=False,
+                               cookies=login_cookies(_util.now()))
         log.debug("2FA location: %s", r.headers.get("Location"))
+        url = r.headers.get("Location")
+        if url and url.startswith("https://www.messenger.com/login/auth_token/"):
+            return url
         url, data = find_form_request(await r.text())
 
     log.info("Starting Facebook checkup flow")
-    r = await session.post(url, data=data, allow_redirects=False)
+    r = await session.post(url, data=data, allow_redirects=False,
+                           cookies=login_cookies(_util.now()))
     log.debug("2FA location: %s", r.headers.get("Location"))
 
     url, data = find_form_request(await r.text())
+    if "verification_method" in data:
+        raise _exception.NotLoggedIn(
+            "Your account is locked, and you need to log in using a browser, and verify it there!"
+        )
     if "submit[This was me]" not in data or "submit[This wasn't me]" not in data:
         raise _exception.ParseError("Could not fill out form properly (2)", data=data)
     data["submit[This was me]"] = "[any value]"
     del data["submit[This wasn't me]"]
     log.info("Verifying login attempt")
 
-    r = await session.post(url, data=data, allow_redirects=False)
+    r = await session.post(url, data=data, allow_redirects=False,
+                           cookies=login_cookies(_util.now()))
     log.debug("2FA location: %s", r.headers.get("Location"))
 
     url, data = find_form_request(await r.text())
@@ -196,7 +207,8 @@ async def two_factor_helper(session: aiohttp.ClientSession, r: aiohttp.ClientRes
     data["name_action_selected"] = "save_device"
     log.info("Saving device again")
 
-    r = await session.post(url, data=data, allow_redirects=False)
+    r = await session.post(url, data=data, allow_redirects=False,
+                           cookies=login_cookies(_util.now()))
     log.debug("2FA location: %s", r.headers.get("Location"))
     return r.headers.get("Location")
 
@@ -257,9 +269,9 @@ class Session:
 
     # TODO: Add ability to load previous cookies in here, to avoid 2fa flow
     @classmethod
-    async def login(
-        cls, email: str, password: str, on_2fa_callback: Callable[[], Awaitable[int]] = None
-    ):
+    async def login(cls, email: str, password: str,
+                    on_2fa_callback: Callable[[], Awaitable[int]] = None,
+                    user_agent: Optional[str] = None) -> 'Session':
         """Login the user, using ``email`` and ``password``.
 
         Args:
@@ -268,10 +280,13 @@ class Session:
             on_2fa_callback: Function that will be called, in case a two factor
                 authentication code is needed. This should return the requested code.
 
-                Only tested using SMS, might not work with authentication applications.
+                Tested using SMS and authentication applications. If you have both
+                enabled, you might not receive an SMS code, and you'll have to use the
+                authentication application.
 
                 Note: Facebook limits the amount of codes they will give you, so if you
                 don't receive a code, be patient, and try again later!
+            user_agent: The user agent to send to Facebook
 
         Example:
             >>> import fbchat
@@ -287,7 +302,7 @@ class Session:
             >>> session.user.id
             "1234"
         """
-        session = session_factory()
+        session = session_factory(user_agent=user_agent)
 
         data = {
             # "jazoest": "2754",
@@ -311,6 +326,7 @@ class Session:
                 "https://www.messenger.com/login/password/",
                 data=data,
                 allow_redirects=False,
+                cookies=login_cookies(_util.now()),
             )
         except aiohttp.ClientError as e:
             _exception.handle_requests_error(e)
@@ -335,18 +351,18 @@ class Session:
             if not url.startswith("https://www.facebook.com/checkpoint/start/"):
                 raise _exception.ParseError("Failed 2fa flow (1)", data=url)
 
-            r = await session.get(url, allow_redirects=False)
+            r = await session.get(url, allow_redirects=False, cookies=login_cookies(_util.now()))
             url = r.headers.get("Location")
             if not url or not url.startswith("https://www.facebook.com/checkpoint/"):
                 raise _exception.ParseError("Failed 2fa flow (2)", data=url)
 
-            r = await session.get(url, allow_redirects=False)
+            r = await session.get(url, allow_redirects=False, cookies=login_cookies(_util.now()))
             url = await two_factor_helper(session, r, on_2fa_callback)
 
             if not url.startswith("https://www.messenger.com/login/auth_token/"):
                 raise _exception.ParseError("Failed 2fa flow (3)", data=url)
 
-            r = await session.get(url, allow_redirects=False)
+            r = await session.get(url, allow_redirects=False, cookies=login_cookies(_util.now()))
             url = r.headers.get("Location")
 
         if url != "https://www.messenger.com/":
@@ -402,7 +418,7 @@ class Session:
             )
 
     @classmethod
-    async def _from_session(cls, session):
+    async def _from_session(cls, session: aiohttp.ClientSession) -> Optional['Session']:
         # TODO: Automatically set user_id when the cookie changes in the session
         user_id = get_user_id(session)
 
@@ -445,7 +461,8 @@ class Session:
         return {key: morsel.value for key, morsel in cookie.items()}
 
     @classmethod
-    async def from_cookies(cls, cookies: Mapping[str, str]):
+    async def from_cookies(cls, cookies: Mapping[str, str], user_agent: Optional[str] = None
+                           ) -> 'Session':
         """Load a session from session cookies.
 
         Args:
@@ -456,7 +473,7 @@ class Session:
             >>> # Store cookies somewhere, and then subsequently
             >>> session = fbchat.Session.from_cookies(cookies)
         """
-        session = session_factory()
+        session = session_factory(user_agent=user_agent)
 
         if isinstance(cookies, BaseCookie):
             cookie = cookies
@@ -522,7 +539,7 @@ class Session:
         return await self._post("/api/graphqlbatch/", data, as_graphql=True)
 
     async def _do_send_request(self, data):
-        now = datetime.datetime.utcnow()
+        now = _util.now()
         offline_threading_id = _util.generate_offline_threading_id()
         data["client"] = "mercury"
         data["author"] = "fbid:{}".format(self._user_id)
